/***********************************************************************
 * Code listing from "Advanced Linux Programming," by CodeSourcery LLC  *
 * Copyright (C) 2001 by New Riders Publishing                          *
 * See COPYRIGHT for license information.                               *
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <signal.h>
#include "bin_sem.h"

/* Create specified number of semaphores */
int semaphore_create(key_t sem_key, int num_semaphores) {
	int semid;

	/* create new semaphore set of semaphores */
	if ((semid = semget(sem_key, num_semaphores, IPC_CREAT | 0666)) < 0) {
		perror("error in creating semaphore");
		exit(1);
	}
	return semid;
}

/* Attach specified number of semaphores - hlee*/
int semaphore_attach(key_t sem_key, int num_semaphores) {
	int semid;

	/* allocate new semaphore set of semaphores */
	if ((semid = semget(sem_key, num_semaphores, 0666)) < 0) {
		perror("error in attaching semaphore");
		exit(1);
	}
	return semid;
}

void semaphore_setval(int semid, int index, int value) { /* procedure to initialize specified semaphore to given value */
	if (semctl(semid, index, SETVAL, value) < 0) {
		perror("error in setting semaphore value");
		exit(1);
	}
}

void semaphore_remove(int semid){
	/* Remove the semaphore from the system and destroy the set of
	 semaphores and data structure associated with it. */
	if (semctl(semid, 0, IPC_RMID) < 0) {
		perror("error in removing semaphore from the system");
		exit(1);
	}
}

/* procedure to perform a P or wait operation on a semaphore of given index */
void semaphore_wait(int semid, int index) {
	struct sembuf sops[1]; /* only one semaphore operation to be executed */

	sops[0].sem_num = index;/* define operation on semaphore with given index */
	sops[0].sem_op = -1; /* subtract 1 to value for P operation */
	sops[0].sem_flg = 0; /* type "man semop" in shell window for details */

	if (semop(semid, sops, 1) == -1) {
		perror("error in semaphore operation(sem_op = -1)");
		exit(1);
	}
}

/* procedure to perform a V or signal operation on semaphore of given index */
void semaphore_post(int semid, int index) {
	struct sembuf sops[1]; /* define operation on semaphore with given index */

	sops[0].sem_num = index;/* define operation on semaphore with given index */
	sops[0].sem_op = 1; /* add 1 to value for V operation */
	sops[0].sem_flg = 0; /* type "man semop" in shell window for details */

	if (semop(semid, sops, 1) == -1) {
		perror("error in semaphore operation(sem_op = 1)");
		exit(1);
	}
}

/* procedure to perform a "wait-for-zero" operation on a semaphore of given index */
void semaphore_waitforzero(int semid, int index) {
	struct sembuf sops[1]; /* only one semaphore operation to be executed */

	sops[0].sem_num = index;/* define operation on semaphore with given index */
	sops[0].sem_op = 0; /* "wait-for-zero" operation */
	sops[0].sem_flg = 0; /* type "man semop" in shell window for details */

	//printf("[syncmgr][bin_sem.c] semaphore_waitforzero(): (-)semop()\n");
	if (semop(semid, sops, 1) == -1) {
		perror("error in semaphore operation(sem_op = 0)");
		exit(1);
	}
	//printf("[syncmgr][bin_sem.c] semaphore_waitforzero(): (+)semop()\n");
}

void semaphore_waitforzero_maskallsig(int semid, int index) {
	/* "wait-for-zero" operation */
	sigset_t newmask, oldmask;
	sigfillset(&newmask);
	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
		perror("SIG_BLOCK error");

	semaphore_waitforzero(semid, index);

	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
		perror("SIG_SETMASK error");
}





// The following code is not used for the project. Just stay here for the future. - by hlee

//int *readcount;	// readcount for readers/writers semaphore
//int *writecount; // writecount for readers/writers semaphore
//
//int shmid_readcount;
//int shmid_writecount;

///* Initialize a binary semaphore with a value of 1. */
//int binary_semaphore_initialize(int semid) {
//	union semun argument;
//	unsigned short values[1];
//	values[0] = 1;
//	argument.array = values;
//	return semctl(semid, 0, SETALL, argument);
//}
//
///* Obtain a binary semaphore’s ID, allocating if necessary.  */
//int binary_semaphore_allocation(key_t key, int sem_flags) {
//	return semget(key, 1, sem_flags);
//}
//
///* Deallocate a binary semaphore. All users must have finished their use.
// Returns -1 on failure. */
//int binary_semaphore_deallocate(int semid) {
//	union semun ignored_argument;
//	return semctl(semid, 1, IPC_RMID, ignored_argument);
//}
//
///* Wait on a binary semaphore.  Block until the semaphore value is
// positive, then decrement it by one.  */
//int binary_semaphore_wait(int semid) {
//	struct sembuf operations[1];
//	/* Use the first (and only) semaphore.  */
//	operations[0].sem_num = 0;
//	/* Decrement by 1.  */
//	operations[0].sem_op = -1;
//	/* Permit undo'ing.  */
//	operations[0].sem_flg = SEM_UNDO;
//
//	return semop(semid, operations, 1);
//}
//
///* Post to a binary semaphore: increment its value by one.  This
// returns immediately.  */
//int binary_semaphore_post(int semid) {
//	struct sembuf operations[1];
//	/* Use the first (and only) semaphore.  */
//	operations[0].sem_num = 0;
//	/* Increment by 1.  */
//	operations[0].sem_op = 1;
//	/* Permit undo'ing.  */
//	operations[0].sem_flg = SEM_UNDO;
//
//	return semop(semid, operations, 1);
//}


//int readerswriters_create(key_t sem_key){
//	return semaphore_create(sem_key, 5);
//}
//
//int readerswriters_attach(key_t sem_key, key_t key_readcount, key_t key_writecount){
//	/* Create shared memory for readcount */
//	if ((shmid_readcount = shmget(key_readcount, sizeof(*readcount),
//			0666)) < 0) {
//		perror("[readerswriters_attach()] shmget-readcount");
//		exit(2);
//	}
//	if ((readcount = (int *) shmat(shmid_readcount, NULL, 0))
//			== (void *) -1) {
//		perror("shmat-readcount");
//		exit(2);
//	}
//
//	/* Create shared memory for readcount */
//	if ((shmid_writecount = shmget(key_writecount, sizeof(*writecount),
//			0666)) < 0) {
//		perror("[readerswriters_attach()] shmget-writecount");
//		exit(2);
//	}
//	if ((writecount = (int *) shmat(shmid_writecount, NULL, 0))
//			== (void *) -1) {
//		perror("shmat-readcount");
//		exit(2);
//	}
//
//	return semaphore_attach(sem_key, 5);
//}
//
//void readerswriters_detach(void){
//	/* Detach is not needed for semaphore */
//
//	/* Detach shared momory */
//	shmdt(readcount);
//	shmdt(writecount);
//}
//
//void readerswriters_remove(int semid){
//	/* Remove semaphore */
//	semaphore_remove(semid);
//
//	/* Remove shared momory */
//	shmdt(readcount);
//	shmctl(shmid_readcount, IPC_RMID, 0);
//	shmdt(writecount);
//	shmctl(shmid_readcount, IPC_RMID, 0);
//}
//
//void readerswriters_initialize(int semid, key_t key_readcount, key_t key_writecount){
//	/* Initialize semaphores */
//	semaphore_setval(semid, RSEM, 1);
//	semaphore_setval(semid, WSEM, 1);
//	semaphore_setval(semid, XSEM, 1);
//	semaphore_setval(semid, YSEM, 1);
//	semaphore_setval(semid, ZSEM, 1);
//
//	/* Create shared memory for readcount */
//	if ((shmid_readcount = shmget(key_readcount, sizeof(*readcount),
//			IPC_CREAT | 0666)) < 0) {
//		perror("[readerswriters_initialize()] shmget-readcount");
//		exit(2);
//	}
//	if ((readcount = (int *) shmat(shmid_readcount, NULL, 0))
//			== (void *) -1) {
//		perror("shmat-readcount");
//		exit(2);
//	}
//
//	/* Create shared memory for readcount */
//	if ((shmid_writecount = shmget(key_writecount, sizeof(*writecount),
//			IPC_CREAT | 0666)) < 0) {
//		perror("[readerswriters_initialize()] shmget-writecount");
//		exit(2);
//	}
//	if ((writecount = (int *) shmat(shmid_writecount, NULL, 0))
//			== (void *) -1) {
//		perror("shmat-readcount");
//		exit(2);
//	}
//
//	/* Initialize *readcount and *writecount */
//	(*readcount) = 0;
//	(*writecount) = 0;
//}
//
//void readers_wait(int semid) {
//	sigset_t newmask, oldmask;
//	sigfillset(&newmask);
//	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
//		perror("SIG_BLOCK error");
//
//	semaphore_wait(semid, ZSEM);
//	semaphore_wait(semid, RSEM);
//	semaphore_wait(semid, XSEM);
//	(*readcount)++;
////	printf("[syncmgr][bin_sem.c] readers_wait(): "
////			"*readcount = %d, *writecount = %d\n", *readcount, *writecount);
//	if ((*readcount) == 1)
//		semaphore_wait(semid, WSEM);
//	semaphore_post(semid, XSEM);
//	semaphore_post(semid, RSEM);
//	semaphore_post(semid, ZSEM);
//
//	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
//		perror("SIG_SETMASK error");
//}
//
//void readers_post(int semid) {
//	sigset_t newmask, oldmask;
//	sigfillset(&newmask);
//	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
//		perror("SIG_BLOCK error");
//
//	semaphore_wait(semid, XSEM);
//	(*readcount)--;
////	printf("[syncmgr][bin_sem.c] readers_post(): "
////			"*readcount = %d, *writecount = %d\n", *readcount, *writecount);
//	if ((*readcount) == 0)
//		semaphore_post(semid, WSEM);
//	semaphore_post(semid, XSEM);
//
//	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
//		perror("SIG_SETMASK error");
//}
//
//void writers_wait(int semid) {
//	sigset_t newmask, oldmask;
//	sigfillset(&newmask);
//	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
//		perror("SIG_BLOCK error");
//
//	semaphore_wait(semid, YSEM);
//	(*writecount)++;
////	printf("[syncmgr][bin_sem.c] writers_wait(): "
////			"*readcount = %d, *writecount = %d\n", *readcount, *writecount);
//	if ((*writecount) == 1) {
//		semaphore_wait(semid, RSEM);
//	}
//	semaphore_post(semid, YSEM);
//	semaphore_wait(semid, WSEM);
//
//	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
//		perror("SIG_SETMASK error");
//}
//
//void writers_post(int semid) {
//	sigset_t newmask, oldmask;
//	sigfillset(&newmask);
//	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
//		perror("SIG_BLOCK error");
//
//	semaphore_post(semid, WSEM);
//	semaphore_wait(semid, YSEM);
//	(*writecount)--;
////	printf(
////			"[syncmgr][bin_sem.c] writers_post(): *readcount = %d, *writecount = %d\n",
////			*readcount, *writecount);
//	if ((*writecount) == 0)
//		semaphore_post(semid, RSEM);
//	semaphore_post(semid, YSEM);
//
//	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
//		perror("SIG_SETMASK error");
//}

